class DigitalTwin:
    """
    Class that represents a digital twin that
    uses a mathematical prediction model to
    analyze an exercise in the real time.\n
    Args:
        predicted_heart_rate (int):
            a predicted heart rate of an interval
            (speed or rest segment) in beats per minute
        duration (int):
            duration of an interval in minutes
    """
    def __init__(
        self,
        proposed_heart_rate: int,
        duration: int,
        tick_time: float
    ) -> None:
        """
        Initialization method of the DigitalTwin class.\n
        Args:
            predicted_heart_rate (int):
                a predicted heart rate of an interval
                (speed or rest segment) in beats per minute
            duration (int):
                duration of an interval in minutes
            tick_time (float):
                a tick time in seconds
        """
        self.proposed_heart_rate = proposed_heart_rate
        self.predicted_duration = duration
        self.current_heart_rate = 0
        self.current_duration = 0
        self.expected_trackpoints = (1000 * duration) // tick_time

    def read_control_data(self) -> tuple:
        """
        Reading the current heart rate and the interval duration.\n
        Returns:
            tuple[int, float]: current heart rate, current duration
        """
        return self.current_heart_rate, self.current_duration

    def calculate_prediction(
        self,
        trimp_delta: float,
        average_heart_rate: float,
        proposed_heart_rate: int,
        time_delta: float,
        expected_trackpoints: int
    ) -> float:
        """
        Calculating the predicted heart rate.\n
        Args:
            trimp_delta (float):
                the calculated TRIMP difference
            average_heart_rate (float):
                the average heart rate of the interval in beats per minute
            proposed_heart_rate (int):
                the proposed heart rate of the interval in beats per minute
            time_delta (float):
                the time difference in seconds
            expected_trackpoints (int):
                the number of expected trackpoints
        Returns:
            float: the proposed heart rate
        """
        if trimp_delta < 0:
            self.minus += trimp_delta
        else:
            self.plus += trimp_delta

        if average_heart_rate < proposed_heart_rate:
            return proposed_heart_rate
        else:
            return (
                proposed_heart_rate +
                (self.minus - self.plus) / (expected_trackpoints * time_delta)
            )

    def predict_heart_rate(self) -> None:
        """
        Digital twin algorithm that monitors the athlete
        performance in the real time and calculates the
        predicted heart rate.
        """
        n = 1
        current_time = 0
        current_heart_rate = 0
        previous_heart_rate, previous_time = self.read_control_data()
        average_heart_rate = 0
        time_delta = 0
        self.minus, self.plus = 0, 0

        while current_time <= self.predicted_duration - 1:
            # Reading the current heart rate and the duration.
            current_heart_rate, current_time = self.read_control_data()

            # Calculating the average heart rate and the time difference.
            average_heart_rate = (
                average_heart_rate +
                (1 / n) * (self.current_heart_rate - average_heart_rate)
            )

            time_delta = current_time - previous_time
            if time_delta == 0:
                continue

            if self.predicted_duration - 1 <= current_time:
                break

            # Calculating the TRIMP difference.
            trimp_delta = (
                ((current_heart_rate - previous_heart_rate) / 2) * time_delta -
                self.proposed_heart_rate * time_delta
            )

            # Predicted heart rate calculation.
            self.predicted_heart_rate = self.calculate_prediction(
                trimp_delta,
                average_heart_rate,
                self.proposed_heart_rate,
                self.predicted_duration - current_time,
                self.expected_trackpoints
            )
            if self.predicted_heart_rate:
                self.predicted_heart_rate = round(self.predicted_heart_rate)

            n += 1
            previous_time = current_time
            previous_heart_rate = current_heart_rate



from sport_activities_features.hill_identification import HillIdentification


class SendData(object):
    """
    Class for sending data to the server and
    making the analysis of the training.
    """
    @staticmethod
    def send_data(data: list) -> None:
        """
        [WIP]
        Currently, the data is not sent to the API,
        instead, only some features from sport-activities-features
        are called.

        Sending data to the API.\n
        Args:
            data (list[SensorData]):
                training parameters
        """
        altitudes = []
        for trackpoint in data:
            altitudes.append(trackpoint.altitude)

        hill_identification = HillIdentification(altitudes, 30)
        hill_identification.identify_hills()
        hills = hill_identification.return_hills()
        print(hills)



import random
from tcxreader.tcxreader import TCXExercise, TCXReader
import time


class Simulation():
    """
    Implementation of methods for simulating heart rate and GPS data.\n
    Args:
        hr_path (str):
            path to the file with heart rates
        gps_path (str):
            path to the file with GPS data
    Date:
        2021
    License:
        MIT
    """
    def __init__(
        self,
        hr_path='../sensor_data/hr.txt',
        gps_path='../sensor_data/gps.txt'
    ) -> None:
        """
        Initialisation method for Simulation class.\n
        Args:
            hr_path (str):
                path to the file with heart rates
            gps_path (str):
                path to the file with GPS data
        """
        self.hr_path = hr_path
        self.gps_path = gps_path

    def write_hr_to_file(self, hr: int) -> None:
        """
        Write HR simulation data to the text file.\n
        Args:
            hr (int):
                heart rate
        """
        with open(self.hr_path, 'a') as f:
            f.write(str(hr) + '\n')
            print('HR: ' + str(hr))

    def write_gps_to_file(self, lon: float, lat: float, alt: float) -> None:
        """
        Write GPS simulation data to the text file.\n
        Args:
            lon (float):
                longitude on Earth
            lat (float):
                latitude on Earth
            alt (float):
                current altitude
        """
        with open(self.gps_path, 'a') as f:
            f.write(str(lon) + ';' + lat + ';' + str(alt) + ';' + '\n')
            print('Lon: ' + str(lon) + ' Lat: ' + str(lat))

    def simulate_hr(self) -> None:
        """
        Randomly generate heart rate between 150 and 160 beats per minute.
        """
        while True:
            time.sleep(2)
            new_rand = random.randint(150, 160)
            self.write_hr_to_file(new_rand)

    def simulate_gps(self) -> None:
        """
        GPS simulation.
        """
        tcx_reader = TCXReader()
        file_location = '../development/datasets/15.tcx'
        """
            TCX file test2.tcx was taken from the following dataset:
            S. Rauter, I. Jr. Fister, I. Fister.

            A collection of sport activity files
            for data analysis and data mining.

            Technical report 0201, University of Ljubljana
            and University of Maribor, 2015.
        """
        data: TCXExercise = tcx_reader.read(file_location)

        lon = []
        lat = []

        for i in range(len(data.trackpoints)):
            lat.append(str(data.trackpoints[i].latitude))
            lon.append(str(data.trackpoints[i].longitude))

        indexed = 0
        while True:
            time.sleep(1)
            self.write_gps_to_file(
                lon[indexed],
                lat[indexed],
                random.randint(1, 3)
            )
            indexed = indexed + 1



import time

# It should work on ARM architectures only.
try:
    import adafruit_gps
    import serial
except BaseException:
    pass


class GpsSensor:
    """
    Class for working with GPS sensor.\n
    Args:
        gps_path (str):
            path to file for storing GPS data
    """
    def __init__(self, gps_path: str = 'sensor_data/gps.txt') -> None:
        """
        Initialisation method for GpsSensor class.\n
        Args:
            gps_path (str):
                path to file for storing gps data
        """
        self.gps_path = gps_path

    def write_gps_data_to_file(
        self,
        longitude: float,
        latitude: float,
        altitude: float
    ) -> None:
        """
        Method for writing GPS data to text file.\n
        Args:
            longitude (float):
                longitude on Earth
            latitude (float):
                latitude on Earth
            altitude (float):
                current altitude
        """
        output = str(longitude) + ';' + str(latitude) + ';' + str(altitude)
        with open(self.gps_path, 'a') as f:
            f.write(output + '\n')

    def get_gps_data(self) -> None:
        """
        Method for listening the channel for obtaining GPS data from sensor.\n
        Note: Example is based on source code from
              https://github.com/adafruit/Adafruit_CircuitPython_GPS
        """
        uart = serial.Serial('/dev/serial0', baudrate=9600, timeout=10)
        gps = adafruit_gps.GPS(uart, debug=False)
        gps.send_command(b'PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0')
        gps.send_command(b'PMTK220,1000')
        last_print = time.monotonic()

        while True:
            gps.update()

            current = time.monotonic()
            if current - last_print >= 1.0:
                last_print = current
                if not gps.has_fix:
                    print('Waiting for fix...')
                    continue

                LATITUDE = gps.latitude
                LONGITUDE = gps.longitude
                ALTITUDE = gps.altitude_m

                if (
                    (LATITUDE is not None) and
                    (LONGITUDE is not None) and
                    (ALTITUDE is not None)
                ):
                    self.write_gps_data_to_file(LONGITUDE, LATITUDE, ALTITUDE)



from ast_monitor.model import AST
from ast_monitor.classes import SensorData, Interval
from ast_monitor.digital_twin import DigitalTwin
from ast_monitor.gps_sensor import GpsSensor
from ast_monitor.hr_sensor import HrSensor
from ast_monitor.simulation import Simulation
from ast_monitor.mainwindow import Ui_MainWindow

__all__ = [
    AST,
    SensorData,
    Interval,
    DigitalTwin,
    GpsSensor,
    HrSensor,
    Simulation,
    Ui_MainWindow
]

__project__ = 'ast_monitor'
__version__ = '0.2.0'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ast_monitor']

package_data = \
{'': ['*'], 'ast_monitor': ['icons/*', 'map/*']}

install_requires = \
['PyQt5',
 'PyQtWebEngine>=5.15.5,<6.0.0',
 'geopy',
 'matplotlib',
 'pyqt-feedback-flow',
 'sport-activities-features>=0.2.18,<0.3.0',
 'tcxreader']

setup_kwargs = {
    'name': 'ast-monitor',
    'version': '0.2.0',
    'description': 'AST-Monitor is a wearable Raspberry Pi computer for cyclists',
    'long_description': '# AST-Monitor --- A wearable Raspberry Pi computer for cyclists\n\n---\n\n[![PyPI Version](https://img.shields.io/pypi/v/ast-monitor.svg)](https://pypi.python.org/pypi/ast-monitor)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/ast-monitor.svg)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/ast-monitor.svg)\n[![Downloads](https://pepy.tech/badge/ast-monitor)](https://pepy.tech/project/ast-monitor)\n[![GitHub license](https://img.shields.io/github/license/firefly-cpp/ast-monitor.svg)](https://github.com/firefly-cpp/AST-Monitor/blob/master/LICENSE)\n![GitHub commit activity](https://img.shields.io/github/commit-activity/w/firefly-cpp/ast-monitor.svg)\n![GitHub contributors](https://img.shields.io/github/contributors/firefly-cpp/ast-monitor.svg)\n\n[![DOI](https://img.shields.io/badge/DOI-10.1109/ISCMI53840.2021.9654817-blue)](https://doi.org/10.1109/ISCMI53840.2021.9654817)\n[![Fedora package](https://img.shields.io/fedora/v/python3-ast-monitor?color=blue&label=Fedora%20Linux&logo=fedora)](https://src.fedoraproject.org/rpms/python-ast-monitor)\n\n<p align="center">\n  <img width="600" src="https://user-images.githubusercontent.com/73126820/171687041-313ee6be-4bde-4c8f-883f-1271a9ba4d85.png" alt="AST-GUI">\n  <img width="600" src="https://user-images.githubusercontent.com/73126820/171687065-04c80ed4-8062-4410-9856-d4c959ab0ab6.png" alt="AST-Map">\n  <img width="600" src="https://user-images.githubusercontent.com/73126820/175985785-479d982d-1a60-46d1-b3d1-927f222d2ae1.png" alt="AST-Intervals">\n</p>\n\nThis repository is devoted to the AST-monitor, i.e., a low-cost and efficient embedded device for monitoring the realization of sports training sessions that are dedicated to monitoring cycling training sessions.\nAST-Monitor is a part of the Artificial Sport Trainer (AST) system. The first bits of AST-monitor were presented in the following [paper](https://arxiv.org/abs/2109.13334).\n\n## Outline of this repository\n\nThis repository presents basic code regarded to GUI. It was ported from the initial prototype to poetry.\n\n## Hardware outline\n\nThe complete hardware part is shown in the figure from which it can be seen that the AST-computer is split into the following pieces:\n\n* a platform with fixing straps that attach to a bicycle,\n* the Raspberry Pi 4 Model B micro-controller with Raspbian OS installed,\n* a five-inch LCD touch screen display,\n* a USB ANT+ stick,\n* Adafruit\'s Ultimate GPS HAT module.\n\n<p align="center">\n  <img width="600" src="https://raw.githubusercontent.com/firefly-cpp/AST-Monitor/main/.github/img/complete_small.JPG" alt="AST-Monitor">\n</p>\n\n\nA Serial Peripheral Interface (SPI) protocol was dedicated to communication between the Raspberry Pi and the GPS peripheral. A specialized USB ANT+ stick was used to capture the HR signal. The screen display was connected using a modified (physically shortened) HDMI cable, while the touch feedback was implemented using physical wires. The computer was powered during the testing phase using the Trust\'s (5 VDC) power bank. The AST-Monitor prototype is still a little bulky, but a more discrete solution is being searched for, including the sweat drainer of the AST.\n\n## Software outline\n\n### Dependencies\n\nList of dependencies:\n\n| Package      | Version    | Platform |\n| ------------ |:----------:|:--------:|\n| PyQt5        | ^5.15.6    | All      |\n| matplotlib   | ^3.5.1     | All      |\n| geopy        | ^2.2.0     | All      |\n| openant        | v0.4     | All      |\n| pyqt-feedback-flow       | ^0.1.0     | All      |\n| tcxreader       | ^0.3.8     | All      |\n| sport-activities-features       | ^0.2.9     | All      |\n\nNote: openant package should be installed manually. Please follow the [official instructions](https://github.com/Tigge/openant). If you use Fedora OS, you can install openant package using the dnf package manager:\n\n```sh\n$ dnf install python-openant\n```\n\nAdditional note: adafruit-circuitpython-gps package must be installed in order to work with the GPS sensor:\n\n```sh\n$ pip install adafruit-circuitpython-gps\n```\n\n## Installation\n\nInstall AST-Monitor with pip:\n\n```sh\n$ pip install ast-monitor\n```\nIn case you want to install directly from the source code, use:\n\n```sh\n$ git clone https://github.com/firefly-cpp/AST-Monitor.git\n$ cd AST-Monitor\n$ poetry build\n$ python setup.py install\n```\n\nTo install AST-Monitor on Fedora Linux, please use:\n\n```sh\n$ dnf install python3-ast-monitor\n```\n\n## Deployment\n\nOur project was deployed on a Raspberry Pi device using Raspbian OS.\n\n### Run AST-Monitor on startup\n\nAdd the following lines in /etc/profile which ensures to run scripts on startup:\n\n```sh\nsudo python3 /home/user/run_example.py\nsudo nohup python3 /home/user/read_hr_data.py  &\nsudo nohup python3 /home/user/read_gps_data.py  &\n```\n## Examples\n\n### Basic run\n\n```python\nfrom PyQt5 import QtCore, QtGui, uic, QtWidgets\nfrom ast_monitor.model import AST\nimport sys\n\n# provide data locations\n\nhr_data = "sensor_data/hr.txt"\ngps_data = "sensor_data/gps.txt"\n\n\nif __name__ == "__main__":\n    app = QtWidgets.QApplication(sys.argv)\n    window = AST(hr_data, gps_data)\n\n    window.show()\n    sys.exit(app.exec_())\n```\n\n\n## License\n\nThis package is distributed under the MIT License. This license can be found online at <http://www.opensource.org/licenses/MIT>.\n\n## Disclaimer\n\nThis framework is provided as-is, and there are no guarantees that it fits your purposes or that it is bug-free. Use it at your own risk!\n\n## References\n\nFister Jr, I., Fister, I., Iglesias, A., Galvez, A., Deb, S., & Fister, D. (2021). On deploying the Artificial Sport Trainer into practice. arXiv preprint [arXiv:2109.13334](https://arxiv.org/abs/2109.13334).\n\nFister Jr, I., Salcedo-Sanz, S., Iglesias, A., Fister, D., Gálvez, A., & Fister, I. (2021). New Perspectives in the Development of the Artificial Sport Trainer. Applied Sciences, 11(23), 11452. DOI: [10.3390/app112311452](https://doi.org/10.3390/app112311452)\n',
    'author': 'Iztok Fister Jr.',
    'author_email': 'iztok@iztok-jr-fister.eu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QHBoxLayout, QLabel, QVBoxLayout
from sport_activities_features.training_loads import BanisterTRIMP


class SensorData:
    """
    Class for storing the captured sensor data.\n
    Args:
        heart_rate (int):
            the current heart rate in beats per minute
        longitude (float):
            the current longitude in degrees
        latitude (float):
            the current latitude in degrees
        time ():
            XXX
        distance (float):
            the current cummulative distance in meters
    """
    def __init__(
        self,
        heart_rate: int,
        longitude: float,
        latitude: float,
        altitude: float,
        time,
        distance: float
    ) -> None:
        """
        Initialization method for SensorData class.\n
        Args:
            heart_rate (int):
                the current heart rate in beats per minute
            longitude (float):
                the current longitude in degrees
            latitude (float):
                the current latitude in degrees
            time ():
                XXX
            distance (float):
                the current cummulative distance in meters
        """
        self.heart_rate = heart_rate
        self.longitude = longitude
        self.latitude = latitude
        self.altitude = altitude
        self.time = time
        self.distance = distance


class Interval:
    """
    Class for storing intervals.\n
    Args:
        speed_duration (int):
            duration of a speed segment of the interval in minutes
        recovery_duration (int):
            duration of a recovery segment of the interval in minutes
        speed_heart_rate (int):
            planned average heart rate during
            a speed segment of the interval
        recovery_heart_rate (int):
            planned average heart rate during
            a recovery segment of the interval
    """
    def __init__(
        self,
        speed_duration: int,
        recovery_duration: int,
        speed_heart_rate: int,
        recovery_heart_rate: int
    ) -> None:
        """
        Initialization method for Interval class.\n
        Args:
            speed_duration (int):
                duration of a speed segment of the interval in minutes
            recovery_duration (int):
                duration of a recovery segment of the interval in minutes
            speed_heart_rate (int):
                planned average heart rate during
                a speed segment of the interval
            recovery_heart_rate (int):
                planned average heart rate during
                a recovery segment of the interval
        """
        self.speed_duration = speed_duration
        self.recovery_duration = recovery_duration
        self.speed_heart_rate = speed_heart_rate
        self.recovery_heart_rate = recovery_heart_rate
        banister = BanisterTRIMP(speed_duration, speed_heart_rate)
        self.planned_TRIMP = banister.calculate_TRIMP()

    def render(self, index: int) -> None:
        """
        Rendering an interval.
        """
        self.lbl_index = QLabel(str(index))
        self.lbl_index.setFont(QFont('Bahnschrift Semilight', 15))

        self.lbl_speed_heart_rate = QLabel(str(self.speed_heart_rate))
        self.lbl_speed_heart_rate.setFont(QFont('Bahnschrift Semilight', 15))
        self.lbl_speed_duration = QLabel(str(self.speed_duration) + ' min')
        self.lbl_speed_duration.setFont(QFont('Bahnschrift Semilight', 15))

        self.lbl_recovery_heart_rate = QLabel(str(self.recovery_heart_rate))
        self.lbl_recovery_heart_rate.setFont(
            QFont('Bahnschrift Semilight', 15)
        )
        self.lbl_recovery_duration = QLabel(
            str(self.recovery_duration) + ' min'
        )
        self.lbl_recovery_duration.setFont(QFont('Bahnschrift Semilight', 15))

        self.lyt_interval = QHBoxLayout()
        self.lyt_interval.addWidget(self.lbl_index)
        self.lyt_interval.addWidget(self.lbl_speed_heart_rate)
        self.lyt_interval.addWidget(self.lbl_speed_duration)
        self.lyt_interval.addWidget(self.lbl_recovery_heart_rate)
        self.lyt_interval.addWidget(self.lbl_recovery_duration)

    @staticmethod
    def render_intervals(intervals: list) -> QVBoxLayout:
        """
        Rendering all intervals from the list.\n
        Args:
            intervals (list):
                list of intervals to be rendered
        Returns:
            QVBoxLayout: layout with intervals
        """
        lyt_training_intervals = QVBoxLayout()
        lyt_training_intervals.setContentsMargins(40, 0, 40, 0)

        # Adding title bar.
        lbl_index = QLabel('Index')
        lbl_index.setFont(QFont('Bahnschrift Semibold', 15))
        lbl_speed_heart_rate = QLabel('Speed HR')
        lbl_speed_heart_rate.setFont(QFont('Bahnschrift Semibold', 15))
        lbl_speed_duration = QLabel('Speed time')
        lbl_speed_duration.setFont(QFont('Bahnschrift Semibold', 15))
        lbl_recovery_heart_rate = QLabel('Rest HR')
        lbl_recovery_heart_rate.setFont(QFont('Bahnschrift Semibold', 15))
        lbl_recovery_duration = QLabel('Rest time')
        lbl_recovery_duration.setFont(QFont('Bahnschrift Semibold', 15))
        lyt_title = QHBoxLayout()
        lyt_title.addWidget(lbl_index)
        lyt_title.addWidget(lbl_speed_heart_rate)
        lyt_title.addWidget(lbl_speed_duration)
        lyt_title.addWidget(lbl_recovery_heart_rate)
        lyt_title.addWidget(lbl_recovery_duration)
        lyt_training_intervals.addLayout(lyt_title)

        index = 1
        for interval in intervals:
            interval.render(index)
            lyt_training_intervals.addLayout(interval.lyt_interval)
            index += 1

        return lyt_training_intervals



from datetime import datetime
import geopy.distance
import json
import os
from pyqt_feedback_flow.feedback import (
    AnimationDirection,
    AnimationType,
    TextFeedback
)
from PyQt5 import (
    QtCore,
    QtWidgets,
    QtWebChannel,
    QtWebEngineWidgets
)
from PyQt5.QtCore import pyqtSlot, Qt, QTimer
from PyQt5.QtWidgets import QLabel, QMessageBox
from threading import Thread

from ast_monitor.classes import Interval, SensorData
from ast_monitor.digital_twin import DigitalTwin
from ast_monitor.mainwindow import Ui_MainWindow
from ast_monitor.send_data import SendData


TICK_TIME = 1000


class AST(QtWidgets.QMainWindow, Ui_MainWindow):
    """
    Main class that represents the interface between computer and athlete.\n
    Args:
        hr_data_path (str):
            path to file for storing HR data
        gps_path (str):
            path to file for storing GPS data
    """
    def __init__(self, hr_data_path: str, gps_data_path: str) -> None:
        """
        Initialisation method for AST class.\n
        Args:
            hr_data_path (str):
                path to file for storing HR data
            gps_path (str):
                path to file for storing GPS data
        """
        self.initialize_GUI()  # GUI initialization.

        # Paths to heart rate and GPS data files.
        self.hr_data_path = hr_data_path
        self.gps_data_path = gps_data_path

        # Structures for storing data and intervals.
        self.time_series = []
        self.planned_intervals = []
        self.intervals = []

        # Current interval data.
        self.current_interval = 0
        self.is_speed_phase = False
        self.interval_average_heart_rate = 0
        self.interval_checkpoints = 0

        #
        self.training_timer = QTimer()

    def initialize_GUI(self) -> None:
        """
        Initialization of the AST-Monitor GUI.
        """
        QtWidgets.QMainWindow.__init__(self, flags=Qt.FramelessWindowHint)
        Ui_MainWindow.__init__(self)
        self.setupUi(self)

        # Map setting.
        self.view = QtWebEngineWidgets.QWebEngineView()
        self.channel = QtWebChannel.QWebChannel()
        self.channel.registerObject("MainWindow", self)
        self.view.page().setWebChannel(self.channel)
        file = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            '../ast_monitor/map/map.html',
        )
        self.view.setUrl(QtCore.QUrl.fromLocalFile(file))
        self.vb_map.addWidget(self.view)

        # Show data in real time.
        self.timer = QTimer()
        self.timer.timeout.connect(self.update_data)
        self.timer.start(1000)

        # Show interval data in real time.
        self.interval_timer = QTimer()
        self.interval_timer.timeout.connect(self.update_interval_data)
        self.interval_timer.start(1000)

        # Shutdown the computer.
        self.btn_shutdown.clicked.connect(self.shutdown_now)

        # On clicked buttons.
        self.tracker = QTimer()
        self.tracker.setInterval(TICK_TIME)
        self.tracker.timeout.connect(self.tick)
        self.do_reset()

        # Interval tick.
        self.interval_tracker = QTimer()
        self.interval_tracker.setInterval(TICK_TIME)
        self.interval_tracker.timeout.connect(self.interval_tick)
        self.do_interval_reset()

        # Menu buttons.
        self.action_about_program.triggered.connect(self.about)
        self.action_license.triggered.connect(self.license)
        self.action_disclaimer.triggered.connect(self.disclaimer)

        # Tracking buttons.
        self.tracking_flag = False
        self.btn_start_tracking.clicked.connect(self.start_tracker)
        self.btn_stop_tracking.clicked.connect(self.end_tracker)

        # Left and right menu buttons.
        self.btn_move_left.clicked.connect(self.menu_left_move)
        self.btn_move_right.clicked.connect(self.menu_right_move)

        # Interval buttons.
        self.interval_flag = False

        # Training button.
        self.btn_load_training.clicked.connect(self.load_training)
        self.btn_start_training.clicked.connect(self.start_training)

    def start_tracker(self) -> None:
        """
        Start of the workout tracking.
        """
        self.tracker.start()
        self.tracking_flag = True

        self.widget_start_stop.setCurrentIndex(1)

        # Show simple notification that workout just started.
        self._feedback = TextFeedback(text='Workout started!')
        self._feedback.show(
            AnimationType.VERTICAL,
            AnimationDirection.UP,
            time=3000
        )

    def end_tracker(self) -> None:
        """
        End of the workout tracking.
        """
        self.tracker.stop()
        self.tracking_flag = False

        self.widget_start_stop.setCurrentIndex(0)

        # Show simple notification that workout just enden.
        self._feedback = TextFeedback(text='Workout ended!')
        self._feedback.show(
            AnimationType.VERTICAL,
            AnimationDirection.DOWN,
            time=3000
        )

    @pyqtSlot()
    def update_data(self) -> None:
        """
        Rendering all the necessary data.
        """
        self.update_current_heart_rate(label=self.lbl_heart_rate)
        self.update_speed(label=self.lbl_speed)

        # Build_time_series in case tracker is enabled.
        if self.tracking_flag:
            self.build_time_series(series=self.time_series)
            self.update_distance(label=self.lbl_distance)

    @pyqtSlot()
    def update_interval_data(self) -> None:
        """
        Rendering all the necessary interval data.
        """
        # Build_time_series in case tracker is enabled.
        if self.interval_flag:
            self.update_current_heart_rate(label=self.lbl_interval_heart_rate)
            self.update_average_heart_rate(
                self.interval_average_heart_rate,
                self.return_current_hr(),
                label=self.lbl_interval_average_heart_rate
            )

    @pyqtSlot()
    def send_data(self) -> None:
        """
        Sending the training data to the API.
        """
        SendData.send_data(self.time_series)

    @pyqtSlot()
    def do_reset(self) -> None:
        """
        Reset of the stopwatch.
        """
        self.track_time = 0
        self.display_stopwatch()

    @pyqtSlot()
    def do_interval_reset(self) -> None:
        """
        Reset of the interval.
        """
        self.interval_time = 0

    def tick(self) -> None:
        """
        Incrementing time and displaying it on the stopwatch on each tick.
        """
        self.track_time += TICK_TIME / 1000
        self.display_stopwatch()

    def interval_tick(self) -> None:
        """
        Incrementing time and displaying it on
        the interval stopwatch on each tick.
        """
        self.interval_time += TICK_TIME / 1000
        self.display_interval_stopwatch()

        try:
            if self.digital_twin:
                self.digital_twin.current_heart_rate = self.return_current_hr()
                self.digital_twin.current_duration = self.interval_time
                self.lbl_interval_proposed_heart_rate.setText(
                    str(self.digital_twin.predicted_heart_rate)
                )
        except Exception:
            pass

    def convert_time_to_hours_minutes_seconds(self, time: int) -> str:
        """
        Converting time in seconds to HH:MM:SS format.\n
        Args:
            time (int):
                time in seconds
        Returns:
            str: time in HH:MM:SS format
        """
        seconds = int(time % 60)
        minutes = int(time / 60)
        hours = int(minutes / 60)

        time = (
            str(hours) +
            ':' +
            (str(minutes) if minutes > 9 else '0' + str(minutes)) +
            ':' +
            (str(seconds) if seconds > 9 else '0' + str(seconds))
        )

        return time

    def display_stopwatch(self) -> None:
        """
        Displaying the stopwatch in the GUI.
        """
        time = self.convert_time_to_hours_minutes_seconds(self.track_time)
        self.lbl_watch.setText(time)

    def display_interval_stopwatch(self) -> None:
        """
        Displaying the stopwatch in the GUI.
        """
        time = self.convert_time_to_hours_minutes_seconds(self.interval_time)
        self.lbl_interval_watch.setText(time)

    def update_current_heart_rate(self, label: QLabel) -> None:
        """
        Calculating and displaying current heart rate.\n
        Args:
            label (QLabel):
                label to be updated
        """
        hr = self.return_current_hr()
        label.setText(str(hr))

    def update_average_heart_rate(
        self,
        previous_heart_rate: int,
        current_heart_rate: int,
        label: QLabel
    ) -> None:
        """
        Calculating and displaying average heart rate
        with a help of a recursive function.\n
        Args:
            previous_heart_rate (int):
                previous average heart rate
            current_heart_rate (int):
                current heart rate
            label (QLabel):
                label to be updated
        """
        if previous_heart_rate == -1:
            return
        elif previous_heart_rate == 0:
            average_heart_rate = current_heart_rate
        else:
            n = self.interval_checkpoints
            avg_hr = previous_heart_rate
            hr = current_heart_rate
            average_heart_rate = int((1 / n) * ((n - 1) * avg_hr + hr))

        label.setText(str(average_heart_rate))
        self.interval_average_heart_rate = average_heart_rate
        self.interval_checkpoints += 1

    def update_speed(self, label: QLabel) -> None:
        """
        Calculating and displaying current speed.\n
        Args:
            label (QLabel):
                label to be updated
        """
        speed = self.calculate_speed()
        label.setText(f'{speed} km/h')

    def update_distance(self, label: QLabel) -> None:
        """
        Calculating and displaying distance.\n
        Args:
            label (QLabel):
                label to be updated
        """
        dist = self.calculate_distance(self.time_series)

        # If there is no distance made yet, 0.00 km is displayed.
        if not dist:
            label.setText('0.00 km')
        else:
            rounded_dist = round(dist, 2)
            dist_str = '{:.2f}'.format(rounded_dist / 1000)
            label.setText(dist_str + ' km')

    def shutdown_now(self) -> None:
        """
        Shutdown of the system.
        """
        os.system('shutdown now -h')

    def get_current_time(self) -> str:
        """
        Getting current time as string.\n
        Returns:
            str: current time (HH:MM:SS)
        """
        now = datetime.now()
        current_time = now.strftime('%H:%M:%S')
        return current_time

    def get_current_timestamp(self) -> datetime:
        """
        Getting current time as datetime.\n
        Returns:
            datetime: current datetime
        """
        now = datetime.now()
        return now

    def build_time_series(self, series: list) -> None:
        """
        Building time series from the collected
        GPS data, HR data and current time.\n
        Args:
            series (list[SensorData]):
                list of series
        """
        HR = self.return_current_hr()
        GPS_LON, GPS_LAT, GPS_ALT = self.return_current_gps()
        TIME = self.get_current_timestamp()
        DIST = self.calculate_distance(series)
        series.append(
            SensorData(HR, GPS_LON, GPS_LAT, GPS_ALT, TIME, DIST)
        )

    def return_current_hr(self) -> int:
        """
        Get current HR data from file.\n
        Returns:
            int: current heart rate
        """
        # If file is empty.
        # Note: should be improved.
        if os.stat(self.hr_data_path).st_size == 0:
            return 0

        with open(self.hr_data_path, 'r') as ins:
            array = []
            for line in ins:
                array.append(int(line.rstrip()))

            # Use data for calculation of average HR.
            final = str(array[-1])

        return int(final)

    def return_current_gps(self) -> tuple:
        """
        Get current GPS data from file.\n
        Returns:
            tuple(float, float, float):
        """
        # If file is empty.
        # Note: should be improved.
        if os.stat(self.gps_data_path).st_size == 0:
            return 0.0, 0.0, 0.0

        with open(self.gps_data_path, 'r') as ins:
            array = []
            for line in ins:
                array.append(line)

            final1 = str(array[-1].rstrip())
            final = final1.split(";")

        return float(final[0]), float(final[1]), float(final[2])

    def calculate_distance(self, series: list) -> float:
        """
        Calculate distance of the workout.\n
        Args:
            series (list[SensorData])
        Returns:
            float: total distance
        """
        if (len(series) < 2):
            return 0.0

        total_distance = 0.0
        for i in range(len(series) - 1):
            coords_1 = (
                series[i + 1].longitude,
                series[i + 1].latitude
            )
            coords_2 = (
                series[i].longitude,
                series[i].latitude
            )
            total_distance = (
                total_distance +
                abs(geopy.distance.geodesic(coords_1, coords_2).m)
            )

        return round(total_distance, 3)

    def calculate_speed(self) -> float:
        """
        Speed calculation in km/h.\n
        Returns:
            float: current speed in km/h
        """
        if len(self.time_series) < 2:
            return 0.0

        prev_stamp = self.time_series[-2]
        cur_stamp = self.time_series[-1]
        speed = (
            3.6 *
            (cur_stamp.distance - prev_stamp.distance) /
            (cur_stamp.time - prev_stamp.time).total_seconds()
        )
        return round(speed, 1)

    def about(self) -> None:
        """
        Show information about application.
        """
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Information)
        msg.setText('Developed by I. Fister Jr., L. Lukač, 2022')
        msg.setWindowTitle('About this application')
        msg.exec_()

    def license(self) -> None:
        """
        Show license information.
        """
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Information)
        msg.setText('Licensed under MIT license!')
        msg.setWindowTitle('License info!')
        msg.exec_()

    def disclaimer(self) -> None:
        """
        Show disclaimer info.
        """
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Information)
        msg.setWindowTitle('Disclaimer')
        msg.setDetailedText(
            'This framework is provided as-is, and there are no guarantees '
            'that it fits your purposes or that it is bug-free. Use it at '
            'your own risk!')
        msg.exec_()

    def menu_left_move(self) -> None:
        """
        Moving left in the menu.
        """
        index = self.stackedWidget.currentIndex()
        if index == 0:
            return

        self.stackedWidget.setCurrentIndex(index - 1)
        self.widget_title.setCurrentIndex(index - 1)

    def menu_right_move(self) -> None:
        """
        Moving right in the menu.
        """
        index = self.stackedWidget.currentIndex()
        if index == self.stackedWidget.count():
            return

        self.stackedWidget.setCurrentIndex(index + 1)
        self.widget_title.setCurrentIndex(index + 1)

    def start_training(self) -> None:
        """
        Manual training start.
        """
        i = self.current_interval
        if i >= len(self.planned_intervals):
            return

        self.start_interval_speed()

        self.training_timer = QTimer()
        self.training_timer.timeout.connect(self.end_interval_speed)
        self.training_timer.start(
            1000 * self.planned_intervals[i].speed_duration
        )

        self.current_interval += 1

    def start_interval_speed(self):
        """
        Starting a speed part of an interval.
        """
        self.interval_tracker.start()
        self.interval_flag = True
        self.interval_average_heart_rate = 0
        self.interval_duration = 0
        self.interval_time = 0.0
        self.is_speed_phase = True
        self.lbl_interval_proposed_heart_rate.setText('-')

        # Launching the digital twin.
        interval = self.planned_intervals[self.current_interval]
        self.digital_twin = DigitalTwin(
            interval.speed_heart_rate,
            interval.speed_duration,
            TICK_TIME
        )
        self.digital_twin_thread = Thread(
            target=self.digital_twin.predict_heart_rate
        )
        self.digital_twin_thread.start()

        # Show simple notification that an interval just started.
        self._feedback = TextFeedback(text='Interval started!')
        self._feedback.show(
            AnimationType.HORIZONTAL,
            AnimationDirection.RIGHT,
            time=3000
        )

    def end_interval_speed(self):
        """
        Ending a speed part of an interval.
        """
        self.interval_time = 0.0
        self.interval_average_heart_rate = -1
        self.interval_checkpoints = 0
        self.lbl_interval_average_heart_rate.setText('-')
        self.lbl_interval_proposed_heart_rate.setText('-')
        self.is_speed_phase = False

        self.start_interval_rest()

    def start_interval_rest(self):
        """
        Starting a rest part of an interval.
        """
        self.lbl_interval_watch.setText('0:00:00')
        self.lbl_interval_proposed_heart_rate.setText('-')

        # Show simple notification that an interval just started.
        self._feedback = TextFeedback(text='Rest started!')
        self._feedback.show(
            AnimationType.HORIZONTAL,
            AnimationDirection.RIGHT,
            time=3000
        )

        interval = self.planned_intervals[self.current_interval]
        self.training_timer = QTimer()
        self.training_timer.timeout.connect(self.end_interval_rest)
        self.training_timer.start(
            1000 * interval.recovery_duration
        )

        # Launching the digital twin.
        del self.digital_twin
        self.digital_twin = DigitalTwin(
            interval.recovery_heart_rate,
            interval.recovery_duration,
            TICK_TIME
        )
        self.digital_twin_thread = Thread(
            target=self.digital_twin.predict_heart_rate
        )
        self.digital_twin_thread.start()

    def end_interval_rest(self):
        """
        Ending a rest part of an interval.
        """
        self.lbl_interval_proposed_heart_rate.setText('-')
        self.training_timer.stop()

        self.start_training()

    def render_intervals_GUI(self) -> None:
        """
        Rendering the read intervals in the GUI.
        """
        # Removing old intervals or text.
        if self.lbl_no_training:
            self.lbl_no_training.deleteLater()
            self.lbl_no_training = None
        old_intervals = self.lyt_training.takeAt(0)
        self.lyt_training.removeItem(old_intervals)

        # Inserting the read intervals.
        self.lyt_training.insertLayout(
            0,
            Interval.render_intervals(self.planned_intervals)
        )

    def load_training(self) -> None:
        """
        Loading training from JSON file.\n
        Args:
            file (str):
                path to the file
        """
        # Currently hard-coded path to the file.
        file = '../development/trainings/training.json'

        # Opening and reading the training data.
        training = {}
        with open(file, 'r') as f:
            training_json = f.read()
            training = json.loads(training_json)

        # Displaying a notification about training load.
        self._feedback = TextFeedback(text='Training loaded successfully')
        self._feedback.show(
            AnimationType.HORIZONTAL,
            AnimationDirection.RIGHT,
            time=3000
        )

        # Adding and rendering the intervals from the training.
        self.planned_intervals = []
        for interval_group in training['interval']:
            for i in range(interval_group['repetitions']):
                interval = Interval(
                    interval_group['speed_duration'],
                    interval_group['recovery_duration'],
                    interval_group['speed_heart_rate'],
                    interval_group['recovery_heart_rate']
                )
                self.planned_intervals.append(interval)

        # self.intervals = list(map(lambda i: i, training['interval']))
        self.render_intervals_GUI()
